print("import: 'cassandra'")
import cassandra

print("import: 'cassandra.cqlengine'")
import cassandra.cqlengine

print("import: 'cassandra.io'")
import cassandra.io

print("import: 'cassandra.numpy_parser'")
import cassandra.numpy_parser

